
#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/platform_device.h>
#include <linux/interrupt.h>
#include <linux/err.h>
#include <linux/i2c.h>
#include <linux/delay.h>
#include <linux/clk.h>

#include <asm/irq.h>
#include <asm/arch/clock.h>
#include <asm/arch/gpio.h>

#include "mxc_i2c_reg.h"

#define DRIVER_NAME "i2c_armadillo5x0"
#define REVISION    "Rev.1 (2007/--/--)"
#define AUTHOR      "Atmark Techno, Inc."
#define DESCRIPTION "Armadillo-5x0 I2C driver"

#undef DEBUG
#if defined(DEBUG)
#define DEBUG_FUNC()         printk(DRIVER_NAME ": %s()\n", __FUNCTION__)
#define DEBUG_INFO(args...)  printk(DRIVER_NAME ": " args)
#else
#define DEBUG_FUNC()
#define DEBUG_INFO(args...)
#endif
#define PRINT_INFO(args...)  printk(KERN_INFO  DRIVER_NAME ": " args)
#define PRINT_DEBUG(args...) printk(KERN_DEBUG DRIVER_NAME ": " args)
#define PRINT_ERR(args...)   printk(KERN_ERR   DRIVER_NAME ": " args)

#define BIT_CYCLE_DELAY	 (20)
#define BIT_TIMEOUT      (HZ/2)

#define DEFAULT_CLKDIV   (0x17)

#define TRANSFER_TIMEOUT (HZ/10)

static int i2c_clk_divider_table[64] = {
	[0x00] = 30,   32,   36,   42,
		 48,   52,   60,   72,
		 80,   88,   104,  128,
		 144,  160,  192,  240,
	[0x10] = 288,  320,  384,  480,
		 576,  640,  768,  960,
		 1152, 1280, 1536, 1920,
		 2304, 2560, 3072, 3840,
	[0x20] = 22,   24,   26,   28,
		 32,   36,   40,   44,
		 48,   56,   64,   72,
		 80,   96,   112,  128,
	[0x30] = 160,  192,  224,  256,
		 320,  384,  448,  512,
		 640,  768,  896,  1024,
		 1280, 1536, 1792, 2048,
};

struct i2c_armadillo5x0_private {
	unsigned long base;
	struct clk *clk;
	int irq;
	int clkdiv;
	wait_queue_head_t queue;
	int transfer_done;
};

static struct i2c_armadillo5x0_private i2c_priv = {
	.base	= IO_ADDRESS(I2C2_BASE_ADDR),
	.irq	= INT_I2C2,
	.clkdiv	= 960,
};

static u16
mxc_i2c_get_clkdiv_val(int clkdiv)
{
	u16 i;
	for (i=0; i<64; i++) {
		if (i2c_clk_divider_table[i] == clkdiv)
			return i;
	}
	PRINT_ERR("not support clkdiv: %d. "
		  "using default clkdiv(%d).\n",
		  clkdiv, i2c_clk_divider_table[DEFAULT_CLKDIV]);
	return DEFAULT_CLKDIV;
}

static int
mxc_i2c_is_busy(struct i2c_adapter *adap)
{
	struct platform_device *pdev;
	struct i2c_armadillo5x0_private *priv;
	u16 status;

	pdev = i2c_get_adapdata(adap);
	priv = platform_get_drvdata(pdev);

	status = readw(priv->base + MXC_I2SR);
	return (status & MXC_I2SR_IBB);
}

static int
mxc_i2c_wait_iif(struct i2c_adapter *adap)
{
	struct platform_device *pdev;
	struct i2c_armadillo5x0_private *priv;
	int ret;

	pdev = i2c_get_adapdata(adap);
	priv = platform_get_drvdata(pdev);

	priv->transfer_done = 0;
	ret = wait_event_interruptible_timeout(priv->queue,
					       priv->transfer_done,
					       adap->timeout);
	if (ret)
		return -ERESTARTSYS;

	if (!priv->transfer_done)
		PRINT_ERR("transfer timedout.\n");   

	return 0;
}

static int
mxc_i2c_send(struct i2c_adapter *adap, u16 val, u16 flag)
{
	struct platform_device *pdev;
	struct i2c_armadillo5x0_private *priv;
	u16 i2cr;
	int ret;
	
	if (flag & ~(MXC_I2CR_MTX | MXC_I2CR_TXAK | MXC_I2CR_RSTA))
		PRINT_ERR("%s(): invalied argment i2cr flag (0x%04x)\n",
			  __FUNCTION__, flag);
	
	pdev = i2c_get_adapdata(adap);
	priv = platform_get_drvdata(pdev);
	
	i2cr = readw(priv->base + MXC_I2CR);
	i2cr &= ~(MXC_I2CR_MTX | MXC_I2CR_TXAK | MXC_I2CR_RSTA);
	i2cr |= flag;
	writew(i2cr, priv->base + MXC_I2CR);
	writew(val, priv->base + MXC_I2DR);
	
	ret = mxc_i2c_wait_iif(adap);
	
	return 0;
}

static int
_mxc_i2c_recv(struct i2c_adapter *adap, u16 flag, int wait)
{
	struct platform_device *pdev;
	struct i2c_armadillo5x0_private *priv;
	u16 i2cr;
	u16 val;
	
	if (flag & ~(MXC_I2CR_MTX | MXC_I2CR_TXAK | MXC_I2CR_RSTA))
		PRINT_ERR("%s(): invalied argment i2cr flag (0x%04x)\n",
			  __FUNCTION__, flag);
	
	pdev = i2c_get_adapdata(adap);
	priv = platform_get_drvdata(pdev);
	
	i2cr = readw(priv->base + MXC_I2CR);
	i2cr &= ~(MXC_I2CR_MTX | MXC_I2CR_TXAK | MXC_I2CR_RSTA);
	i2cr |= flag;
	writew(i2cr, priv->base + MXC_I2CR);
	
	val = readw(priv->base + MXC_I2DR);
	if (wait)
		mxc_i2c_wait_iif(adap);
	
	return (int)val;
}

static inline int
mxc_i2c_recv(struct i2c_adapter *adap, u16 flag)
{
	return _mxc_i2c_recv(adap, flag, 1);
}

static inline int
mxc_i2c_recv_last(struct i2c_adapter *adap, u16 flag)
{
	return _mxc_i2c_recv(adap, flag, 0);
}

static int 
mxc_i2c_stop(struct i2c_adapter *adap)
{
	struct platform_device *pdev;
	struct i2c_armadillo5x0_private *priv;
	u16 i2cr;
	int retry = 100;
	
	pdev = i2c_get_adapdata(adap);
	priv = platform_get_drvdata(pdev);
	
	i2cr = readw(priv->base + MXC_I2CR);
	i2cr &= (MXC_I2CR_IEN | MXC_I2CR_IIEN);
	writew(i2cr, priv->base + MXC_I2CR);
	
	while (mxc_i2c_is_busy(adap) && retry--)
		udelay(3);

	if (retry < 0)
		PRINT_ERR("%s(): TIMEDOUT\n", __FUNCTION__);
	
	return 0;
}

static int
mxc_i2c_start(struct i2c_adapter *adap)
{
	struct platform_device *pdev;
	struct i2c_armadillo5x0_private *priv;
	u16 i2cr;
	int retry = 100;

	pdev = i2c_get_adapdata(adap);
	priv = platform_get_drvdata(pdev);
	
	i2cr = readw(priv->base + MXC_I2CR);
	i2cr |= MXC_I2CR_MSTA;
	writew(i2cr, priv->base + MXC_I2CR);
	
	while (!mxc_i2c_is_busy(adap) && retry--)
		udelay(3);

	if (retry < 0)
		PRINT_ERR("%s(): TIMEDOUT\n", __FUNCTION__);
	
	return 0;
}

static int
mxc_i2c_enable(struct platform_device *pdev)
{
	struct i2c_armadillo5x0_private *priv;
	
	DEBUG_FUNC();

	priv = platform_get_drvdata(pdev);
	
	//mxc_clks_enable(I2C2_CLK);
	clk_enable(priv->clk);
	
	writew(mxc_i2c_get_clkdiv_val(priv->clkdiv), priv->base + MXC_IFDR);
	writew(0x0000, priv->base + MXC_I2SR);
	writew(MXC_I2CR_IEN, priv->base + MXC_I2CR);
	writew((MXC_I2CR_IEN | MXC_I2CR_IIEN), priv->base + MXC_I2CR);
	
	return 0;
}

static int
mxc_i2c_disable(struct platform_device *pdev)
{
	struct i2c_armadillo5x0_private *priv;
	
	DEBUG_FUNC();
	
	priv = platform_get_drvdata(pdev);
	
	writew(0x0000, priv->base + MXC_I2CR);
	
	//mxc_clks_disable(I2C2_CLK);
	clk_disable(priv->clk);
	
	return 0;
}

static irqreturn_t
i2c_armadillo5x0_irq_handler(int irq, void *dev_id, struct pt_regs *regs)
{
	struct platform_device *pdev = dev_id;
	struct i2c_armadillo5x0_private *priv;
	
	priv = platform_get_drvdata(pdev);
	
	writew(0x0000, i2c_priv.base + MXC_I2SR);
	
	priv->transfer_done = 1;
	wake_up_interruptible(&priv->queue);
	
	return IRQ_HANDLED;
}

static int
i2c_armadillo5x0_xfer(struct i2c_adapter *adap, struct i2c_msg *msgs,
		      int num)
{
	struct platform_device *pdev;
	struct i2c_armadillo5x0_private *priv;
	
	int i, j;
	int start_flag = 0;
	u16 val, flag;
	
	DEBUG_FUNC();
	
	pdev = i2c_get_adapdata(adap);
	priv = platform_get_drvdata(pdev);
	
	if (mxc_i2c_is_busy(adap)) {
		PRINT_ERR("Bus busy.\n");
		mxc_i2c_disable(pdev);
		return -EREMOTEIO;
	}

	for (i=0; i<num; i++) {
		if (!start_flag)
			mxc_i2c_start(adap);
		
		flag = MXC_I2CR_MTX;
		
		if ((i > 0) && 
		    (msgs[i].addr == msgs[i-1].addr) && 
		    ((msgs[i].flags & I2C_M_RD) != 
		     (msgs[i-1].flags & I2C_M_RD)))
			flag |= MXC_I2CR_RSTA;
		
		val = ((msgs[i].addr << 1) | 
		       ((msgs[i].flags & I2C_M_RD) ? 1 : 0));
		
		mxc_i2c_send(adap, val, flag);
		
		if (msgs[i].flags & I2C_M_RD) {
			/* Read */
			flag = 0;
			if (msgs[i].len == 1)
				flag = MXC_I2CR_TXAK;

			mxc_i2c_recv(adap, flag); /* DummyRead */
			for (j=0; j<(msgs[i].len - 1); j++) {
				if (j == (msgs[i].len - 1 - 1))
					flag = MXC_I2CR_TXAK;
				val = mxc_i2c_recv(adap, flag);
				msgs[i].buf[j] = val;
				DEBUG_INFO("i2c_recv[%d]: val=%x, flag=%x\n",
					   j, val, flag);
			}
			mxc_i2c_stop(adap);
			start_flag = 0;
			flag = 0;
			val = mxc_i2c_recv_last(adap, flag);
			msgs[i].buf[j] = val;
			DEBUG_INFO("i2c_recv[%d]: val=%x, flag=%x\n",
				   j, val, flag);
		} else {
			/* Write */
			for (j=0; j<msgs[i].len; j++) {
				val = msgs[i].buf[j];
				flag = MXC_I2CR_MTX;
				
				mxc_i2c_send(adap, val, flag);
				DEBUG_INFO("i2c_send[%d]: val=%x, flag=%x\n",
					   j, val, flag);
			}
		}
	}
	mxc_i2c_stop(adap);
	
	return num;
}

static u32
i2c_armadillo5x0_func(struct i2c_adapter *adap)
{
	DEBUG_FUNC();
	return (I2C_FUNC_I2C | I2C_FUNC_SMBUS_EMUL);
}

static int 
i2c_armadillo5x0_client_register(struct i2c_client *client)
{
	DEBUG_FUNC();
	return 0;
}

static int
i2c_armadillo5x0_client_unregister(struct i2c_client *clien)
{
	DEBUG_FUNC();
	return 0;
}

/* module power management */
#if defined(CONFIG_PM)
#include <linux/pm.h>
static int 
i2c_armadillo5x0_suspend(struct platform_device *pdev, pm_message_t state)
{
	DEBUG_FUNC();
	return 0;
}

static int 
i2c_armadillo5x0_resume(struct platform_device *pdev)
{
	DEBUG_FUNC();
	return 0;
}
#endif /* CONFIG_PM */

static struct i2c_algorithm i2c_armadillo5x0_algo = {
	.master_xfer		= i2c_armadillo5x0_xfer,
	.functionality		= i2c_armadillo5x0_func,
};

static struct i2c_adapter i2c_armadillo5x0_adapter = {
	.owner			= THIS_MODULE,
	.name			= "Armadillo-5x0 I2C Adapter",
	.algo			= &i2c_armadillo5x0_algo,
	.client_register	= i2c_armadillo5x0_client_register,
	.client_unregister	= i2c_armadillo5x0_client_unregister,
	.timeout		= TRANSFER_TIMEOUT,
};

static int __devinit 
i2c_armadillo5x0_probe(struct platform_device *pdev)
{
	int ret;
	
	DEBUG_FUNC();

	i2c_priv.clk = clk_get(&pdev->dev, "i2c_clk");
	
	init_waitqueue_head(&i2c_priv.queue);
	platform_set_drvdata(pdev, &i2c_priv);
	
	ret = request_irq(i2c_priv.irq, i2c_armadillo5x0_irq_handler,
			  0, DRIVER_NAME, pdev);
	if (ret < 0) {
		PRINT_ERR("request_irq() failed.\n");
		return ret;
	}

	mxc_i2c_enable(pdev);
	
	i2c_set_adapdata(&i2c_armadillo5x0_adapter, pdev);
	ret = i2c_add_adapter(&i2c_armadillo5x0_adapter);
	if (ret < 0) {
		free_irq(i2c_priv.irq, pdev);
		PRINT_ERR("i2c_add_adapter failed.\n");
		return ret;
	}
	
	return 0;
}

static int 
i2c_armadillo5x0_remove(struct platform_device *pdev)
{
	struct i2c_armadillo5x0_private *priv;
	
	DEBUG_FUNC();
	
	mxc_i2c_disable(pdev);
	i2c_set_adapdata(&i2c_armadillo5x0_adapter, NULL);
	
	priv = platform_get_drvdata(pdev);
	free_irq(priv->irq, pdev);
	i2c_del_adapter(&i2c_armadillo5x0_adapter);
	platform_set_drvdata(pdev, NULL);
	
	return 0;
}

static struct platform_device *i2c_armadillo5x0_device;
static struct platform_driver i2c_armadillo5x0_driver = {
	.probe		= i2c_armadillo5x0_probe,
	.remove		= __devexit_p(i2c_armadillo5x0_remove),
#if defined(CONFIG_PM)
	.suspend	= i2c_armadillo5x0_suspend,
	.resume		= i2c_armadillo5x0_resume,
#endif
	.driver		= {
		.name	= DRIVER_NAME,
	},
};

#if 0
#warning MUX & PAD changed!
#include <asm/arch/mxc_i2c.h>
#include <asm/arch/gpio.h>
#define OUTPUTCONFIG_ALT1 2
#define INPUTCONFIG_ALT1 4
#endif

static int __init 
i2c_armadillo5x0_init(void)
{
	int ret;
	struct platform_device *pdev;
	
	DEBUG_FUNC();
	PRINT_INFO(DESCRIPTION " [" REVISION "], (C) 2007 " AUTHOR "\n");
	
#if 0
#warning MUX & PAD changed!
	mxc_request_iomux(MX31_PIN_CSPI2_MOSI, OUTPUTCONFIG_ALT1,
			  INPUTCONFIG_ALT1);
	mxc_request_iomux(MX31_PIN_CSPI2_MISO, OUTPUTCONFIG_ALT1,
			  INPUTCONFIG_ALT1);
	mxc_iomux_set_pad(MX31_PIN_CSPI2_MOSI, 0x108);
	mxc_iomux_set_pad(MX31_PIN_CSPI2_MISO, 0x108);
#endif
	gpio_i2c_active(1);
	
	ret = platform_driver_register(&i2c_armadillo5x0_driver);
	if (ret < 0)
		goto out_unregister_driver;
	
	pdev = platform_device_register_simple(DRIVER_NAME, -1, NULL, 0);
	i2c_armadillo5x0_device = pdev;
	if (IS_ERR(pdev)) {
		ret = PTR_ERR(pdev);
		goto out_unregister_device;
	}
	
	return 0;
	
 out_unregister_device:
	platform_device_unregister(pdev);
 out_unregister_driver:
	platform_driver_unregister(&i2c_armadillo5x0_driver);    
	
	return ret;
}

static void __exit 
i2c_armadillo5x0_exit(void)
{
	DEBUG_FUNC();
	
	platform_device_unregister(i2c_armadillo5x0_device);
	platform_driver_unregister(&i2c_armadillo5x0_driver);
}

module_init(i2c_armadillo5x0_init);
module_exit(i2c_armadillo5x0_exit);

MODULE_AUTHOR(AUTHOR);
MODULE_DESCRIPTION(DESCRIPTION);
MODULE_LICENSE("GPL");
